from collections import namedtuple
import os
from typing import Iterable

from .errors import AporiaError

ConfigVar = namedtuple("ConfigVar", ["name", "env_var", "default_value", "cast_func"])


def _str_to_bool(value: str) -> bool:
    return value.lower() == "true"


REQUIRED_SETTINGS = [
    ConfigVar("token", "APORIA_TOKEN", None, str),
    ConfigVar("host", "APORIA_HOST", "app.aporia.com", str),
    ConfigVar("environment", "APORIA_ENVIRONMENT", None, str),
]
OPTIONAL_SETTINGS = [
    ConfigVar("port", "APORIA_PORT", 443, int),
    ConfigVar("verbose", "APORIA_VERBOSE", False, _str_to_bool),
    ConfigVar("throw_errors", "APORIA_THROW_ERRORS", False, _str_to_bool),
    ConfigVar("debug", "APORIA_DEBUG", False, _str_to_bool),
    ConfigVar("http_timeout_seconds", "APORIA_HTTP_TIMEOUT_SECONDS", 30, int),
    ConfigVar("verify_ssl", "APORIA_VERIFY_SSL", True, _str_to_bool),
]
ALL_SETTINGS = [*REQUIRED_SETTINGS, *OPTIONAL_SETTINGS]


class Config:
    """Configuration manager."""

    def __init__(self, **kwargs):
        """Initializes a config object.

        Args:
            kwargs: setting overrides.
        """
        settings = self._get_settings_from_env_vars(ALL_SETTINGS)
        overrides = {name: value for name, value in kwargs.items() if value is not None}
        settings.update(overrides)

        self._validate_required_settings(settings, REQUIRED_SETTINGS)

        # This is mostly here for auto-completion and type checks
        # We could set these in a loop using setattr instead
        self.token = settings["token"]
        self.host = settings["host"]
        self.environment = settings["environment"]
        self.port = settings["port"]
        self.verbose = settings["verbose"]
        self.throw_errors = settings["throw_errors"]
        self.debug = settings["debug"]
        self.http_timeout_seconds = settings["http_timeout_seconds"]
        self.verify_ssl = settings["verify_ssl"]

        # Constant values - we keep those here to make overriding them in the tests easier.
        # speed and our expected rate for inserting a prediction
        self.queue_batch_size = 4000
        self.queue_flush_interval = 5
        self.queue_max_size = 500000

        self._validate_environment(self.environment)

    @staticmethod
    def _get_settings_from_env_vars(requested_settings: Iterable[ConfigVar]) -> dict:
        settings = {}

        for config_var in requested_settings:
            value = os.environ.get(config_var.env_var)
            if value is not None:
                settings[config_var.name] = config_var.cast_func(value)
            elif config_var.default_value is not None:
                settings[config_var.name] = config_var.default_value

        return settings

    @staticmethod
    def _validate_required_settings(settings: dict, required_settings: Iterable[ConfigVar]):
        for config_var in required_settings:
            if config_var.name not in settings:
                raise AporiaError(
                    "Required setting '{}' was not defined - pass it to aporia.init or "
                    "define the APORIA_{} environment variable.".format(
                        config_var.name, config_var.env_var
                    )
                )

    @staticmethod
    def _validate_environment(environment: str):
        if len(environment) == 0:
            raise AporiaError("Environment string must be non-empty.")



from .model_versions_api import delete_model_version, get_model_versions, ModelVersion
from .monitor_api import create_monitor, delete_monitor, MonitorType, run_all_monitors

__all__ = [
    "create_monitor",
    "delete_monitor",
    "run_all_monitors",
    "MonitorType",
    "ModelVersion",
    "get_model_versions",
    "delete_model_version",
]



import logging
from typing import Any, Dict, List, Optional, Union

from aporia.core.http_client import HttpClient
from aporia.core.logging_utils import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)


class FieldTrainingData:
    """Field training data."""

    def __init__(
        self,
        field_name: str,
        key: Optional[str] = None,
        bins: Optional[List[Union[float, int, str, bool]]] = None,
        counts: Optional[List[int]] = None,
        min: Optional[float] = None,
        max: Optional[float] = None,
        sum: Optional[float] = None,
        median: Optional[float] = None,
        average: Optional[float] = None,
        std: Optional[float] = None,
        variance: Optional[float] = None,
        num_samples: Optional[int] = None,
        num_missing_values: Optional[int] = None,
        num_posinf_values: Optional[int] = None,
        num_neginf_values: Optional[int] = None,
        num_unique_values: Optional[int] = None,
        num_zero_values: Optional[int] = None,
    ):
        """Initializes a FieldTrainingData object.

        Args:
            field_name: Field name
            key: Key for dict fields.
            bins: Histogram bin edges.
            counts: Hitsogram.
            min: Minimum value.
            max: Maximum Value.
            sum: Sum of all values.
            median: Median values.
            average: Average value.
            std: Standard deviation value.
            variance: Variance value.
            num_samples: Number of data samples.
            num_missing_values: Number of missing values.
            num_posinf_values: Number of positive infinite values.
            num_neginf_values: Number of negative infinite values.
            num_unique_values: Number of unique values.
            num_zero_values: Number of zero values.
        """
        self.field_name = field_name
        self.key = key
        self.bins = bins
        self.counts = counts
        self.min = min
        self.max = max
        self.sum = sum
        self.median = median
        self.average = average
        self.std = std
        self.variance = variance
        self.num_samples = num_samples
        self.num_missing_values = num_missing_values
        self.num_posinf_values = num_posinf_values
        self.num_neginf_values = num_neginf_values
        self.num_unique_values = num_unique_values
        self.num_zero_values = num_zero_values

    def serialize(self) -> dict:
        """Serializes the field training data to a dict.

        Returns:
            Serialized training data.
        """
        return {
            "fieldName": self.field_name,
            "key": self.key,
            "bins": self.bins,
            "counts": self.counts,
            "min": self.min,
            "max": self.max,
            "sum": self.sum,
            "median": self.median,
            "average": self.average,
            "std": self.std,
            "variance": self.variance,
            "numSamples": self.num_samples,
            "numMissingValues": self.num_missing_values,
            "numPosinfValues": self.num_posinf_values,
            "numNeginfValues": self.num_neginf_values,
            "numUniqueValues": self.num_unique_values,
            "numZeroValues": self.num_zero_values,
        }


async def log_training_data(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    features: List[FieldTrainingData],
    labels: List[FieldTrainingData],
    raw_inputs: Optional[List[FieldTrainingData]] = None,
):
    """Reports training data.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Mode version
        features: Training set features.
        labels: Training set labels.
        raw_inputs: Training set raw inputs.
    """
    query = """
        mutation LogTrainingSet(
            $modelId: String!,
            $modelVersion: String!,
            $features: [FieldTrainingData]!
            $labels: [FieldTrainingData]!
            $rawInputs: [FieldTrainingData]
        ) {
            logTrainingSet(
                modelId: $modelId,
                modelVersion: $modelVersion,
                features: $features
                labels: $labels
                rawInputs: $rawInputs
            ) {
                warnings
            }
        }
    """

    serialized_raw_inputs = None
    if raw_inputs is not None:
        serialized_raw_inputs = [field_data.serialize() for field_data in raw_inputs]

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "features": [field_data.serialize() for field_data in features],
        "labels": [field_data.serialize() for field_data in labels],
        "rawInputs": serialized_raw_inputs,
    }

    result = await http_client.graphql(query, variables)
    for warning in result["logTrainingSet"]["warnings"]:
        logger.warning(warning)


async def log_test_data(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    features: List[FieldTrainingData],
    predictions: List[FieldTrainingData],
    labels: List[FieldTrainingData],
    raw_inputs: Optional[List[FieldTrainingData]] = None,
):
    """Reports test data.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Mode version
        features: Test set features.
        predictions: Test set features.
        labels: Test set labels.
        raw_inputs: Test set raw inputs.
    """
    query = """
        mutation LogTestSet(
            $modelId: String!,
            $modelVersion: String!,
            $features: [FieldTrainingData]!
            $predictions: [FieldTrainingData]!
            $labels: [FieldTrainingData]!
            $rawInputs: [FieldTrainingData]
        ) {
            logTestSet(
                modelId: $modelId,
                modelVersion: $modelVersion,
                features: $features
                predictions: $predictions
                labels: $labels
                rawInputs: $rawInputs
            ) {
                warnings
            }
        }
    """

    serialized_raw_inputs = None
    if raw_inputs is not None:
        serialized_raw_inputs = [field_data.serialize() for field_data in raw_inputs]

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "features": [field_data.serialize() for field_data in features],
        "predictions": [field_data.serialize() for field_data in predictions],
        "labels": [field_data.serialize() for field_data in labels],
        "rawInputs": serialized_raw_inputs,
    }

    result = await http_client.graphql(query, variables)
    for warning in result["logTestSet"]["warnings"]:
        logger.warning(warning)


async def log_training_sample_data(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    features: List[Dict[str, Any]],
    labels: List[Any],
    raw_inputs: Optional[List[Any]] = None,
):
    """Reports training sample.

    Args:
        http_client: Http client
        model_id: Model ID.
        model_version: Mode version.
        features: List of features.
        labels: List of labels.
        raw_inputs: List of raw inputs.
    """
    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/training_sample",
        data={"features": features, "labels": labels, "raw_inputs": raw_inputs},
    )



from typing import Any, Optional, Union

from aporia.core.errors import AporiaError
from aporia.inference.inference_model import InferenceModel

try:
    import numpy
    import pandas

    import aporia.training

    ndarray = numpy.ndarray
    DataFrame = pandas.DataFrame
    TRAINING_ENABLED = True

except ImportError:
    ndarray = Any  # type: ignore
    DataFrame = Any
    TRAINING_ENABLED = False

try:
    import numpy
    import pandas
    import pyspark.sql

    import aporia.pyspark

    PySparkDataFrame = pyspark.sql.DataFrame
    PYSPARK_ENABLED = True

except ImportError:
    PySparkDataFrame = Any  # type: ignore
    PYSPARK_ENABLED = False


class Model(InferenceModel):
    """Model object for logging model events."""

    def __init__(self, model_id: str, model_version: str):
        """Initializes a model object.

        Args:
            model_id: Model identifier, as received from the Aporia dashboard.
            model_version: Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
        """
        super().__init__(model_id=model_id, model_version=model_version)

        self._training = None
        if TRAINING_ENABLED:
            self._training = aporia.training.TrainingModel(
                model_id=model_id, model_version=model_version
            )

        self._pyspark_inference = None
        if PYSPARK_ENABLED:
            self._pyspark_inference = aporia.pyspark.PySparkInferenceModel(inference_model=self)

        self._pyspark_training = None
        if PYSPARK_ENABLED and self._training is not None:
            self._pyspark_training = aporia.pyspark.PySparkTrainingModel(
                training_model=self._training
            )

    def __getstate__(self) -> dict:
        """Returns picklable representation of the Model object."""
        return {
            "model_id": self.model_id,
            "model_version": self.model_version,
        }

    def __setstate__(self, state: dict):
        """Loads Model object from pickled state.

        Args:
            state: Pickled state
        """
        self.__init__(  # type: ignore
            model_id=state["model_id"],
            model_version=state["model_version"],
        )

    def log_training_set(
        self,
        features: Union[DataFrame, ndarray],  # type: ignore
        labels: DataFrame,  # type: ignore
        raw_inputs: Optional[DataFrame] = None,  # type: ignore
        sample: bool = True,
    ):
        """Logs aggregations of the whole set and logs a sample of the data.

        Args:
            features: Training set features
            labels: Training set labels
            raw_inputs: Training set raw inputs.
            sample: Whether to log a sample of the data.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._training is None:
                raise AporiaError(
                    short_message="Logging training data failed, Aporia training extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log training data. "
                        "Install it with `pip install aporia[training] and try again."
                    ),
                )

            self._training.log_training_set(
                features=features,
                labels=labels,
                raw_inputs=raw_inputs,
                sample=sample,
            )

    def log_training_sample(
        self,
        features: Union[DataFrame, ndarray],  # type: ignore
        labels: DataFrame,  # type: ignore
        raw_inputs: Optional[DataFrame] = None,  # type: ignore
    ):
        """Logs a sample of the training data.

        Args:
            features: Training set features
            labels: Training set labels
            raw_inputs: Training set raw_inputs

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * labels -> predictions
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._training is None:
                raise AporiaError(
                    short_message="Logging training samples failed, Aporia training extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log training samples. "
                        "Install it with `pip install aporia[training] and try again."
                    ),
                )

            self._training.log_training_data(
                features=features, labels=labels, raw_inputs=raw_inputs
            )

    def log_test_set(
        self,
        features: Union[DataFrame, ndarray],  # type: ignore
        predictions: DataFrame,  # type: ignore
        labels: DataFrame,  # type: ignore
        raw_inputs: Optional[DataFrame] = None,  # type: ignore
        confidences: Optional[ndarray] = None,  # type: ignore
    ):
        """Logs test data.

        Args:
            features: Test set features
            predictions: Test set predictions
            labels: Test set labels
            raw_inputs: Test set raw inputs.
            confidences: Confidence values for the test predictions.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * predictions -> predictions
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._training is None:
                raise AporiaError(
                    short_message="Logging test data failed, Aporia training extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log test data. "
                        "Install it with `pip install aporia[training] and try again."
                    ),
                )

            self._training.log_test_set(
                features=features,
                predictions=predictions,
                labels=labels,
                raw_inputs=raw_inputs,
                confidences=confidences,
            )

    def log_batch_pyspark_raw_inputs(
        self,
        ids: PySparkDataFrame,  # type: ignore
        raw_inputs: PySparkDataFrame,  # type: ignore
    ):
        """Logs raw inputs of multiple predictions.

        Args:
            ids: Prediction identifiers
            raw_inputs: Raw inputs of each prediction

        Notes:
            * The ids dataframe must contain exactly one column
            * The ids and raw_inputs dataframes must have the same number of rows
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark raw inputs failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log raw inputs from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )

            self._pyspark_inference.log_batch_pyspark_raw_inputs(ids=ids, raw_inputs=raw_inputs)

    def log_batch_pyspark_actuals(
        self,
        ids: PySparkDataFrame,  # type: ignore
        actuals: PySparkDataFrame,  # type: ignore
    ):
        """Logs actual values of multiple predictions.

        Args:
            ids: Prediction identifiers
            actuals: Actual prediction results of each prediction

        Notes:
            * The ids dataframe must contain exactly one column
            * The ids and actuals dataframes must have the same number of rows
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark actuals failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log actuals from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )

            self._pyspark_inference.log_batch_pyspark_actuals(ids=ids, actuals=actuals)

    def log_batch_pyspark_prediction(
        self,
        ids: PySparkDataFrame,  # type: ignore
        features: PySparkDataFrame,  # type: ignore
        predictions: PySparkDataFrame,  # type: ignore
        raw_inputs: Optional[PySparkDataFrame] = None,  # type: ignore
        actuals: Optional[PySparkDataFrame] = None,  # type: ignore
    ):
        """Logs multiple predictions.

        Args:
            ids: Prediction identifiers
            features: Values for all of the features in each prediction
            predictions: Prediction results.
            raw_inputs: Raw inputs for each prediction.
            actuals: Actual prediction results of each prediction

        Notes:
            * The ids dataframe must contain exactly one column
            * The ids, features, predictions, raw_inputs and actuals dataframes must
              have the same number of rows
        """
        with self.handle_error("{}"):
            if self._pyspark_inference is None:
                raise AporiaError(
                    short_message="Logging pyspark predictions failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log predictions from pyspark "
                        "dataframes. Install it with `pip install aporia[pyspark] and try again."
                    ),
                )

            self._pyspark_inference.log_batch_pyspark_prediction(
                ids=ids,
                features=features,
                predictions=predictions,
                raw_inputs=raw_inputs,
                actuals=actuals,
            )

    def log_pyspark_training_set(
        self,
        features: PySparkDataFrame,  # type: ignore
        labels: PySparkDataFrame,  # type: ignore
        raw_inputs: Optional[PySparkDataFrame] = None,  # type: ignore
    ):
        """Logs training data from PySpark DataFrames.

        Args:
            features: Training set features
            labels: Training set labels
            raw_inputs: Training set raw inputs.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._pyspark_training is None:
                raise AporiaError(
                    short_message="Logging training data failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia pyspark extension is required to log training data from "
                        "pyspark dataframes. Install it with `pip install aporia[pyspark]` "
                        "and try again."
                    ),
                )

            self._pyspark_training.log_pyspark_training_set(
                features=features, labels=labels, raw_inputs=raw_inputs
            )

    def log_pyspark_test_set(
        self,
        features: PySparkDataFrame,  # type: ignore
        predictions: PySparkDataFrame,  # type: ignore
        labels: PySparkDataFrame,  # type: ignore
        raw_inputs: Optional[PySparkDataFrame] = None,  # type: ignore
    ):
        """Logs test data from PySpark DataFrames.

        Args:
            features: Test set features
            predictions: Test set predictions
            labels: Test set labels
            raw_inputs: Test set raw inputs.

        Notes:
            * Each dataframe corresponds to a field category defined in create_model_version:
                * features -> features
                * predictions -> predictions
                * labels -> predictions
                * raw_inputs -> raw_inputs
            * Each column in the dataframe should match a field defined in create_model_version
                * Missing fields will be handled as missing values
                * Columns that do not match a defined field will be ignored
                * The column name must match the field name
            * This function is blocking and may take a while to finish running.
        """
        with self.handle_error("{}"):
            if self._pyspark_training is None:
                raise AporiaError(
                    short_message="Logging test data failed, Aporia pyspark extension not found",
                    verbose_message=(
                        "The Aporia training extension is required to log test data from "
                        "pyspark dataframes. Install it with `pip install aporia[pyspark]` "
                        "and try again."
                    ),
                )

            self._pyspark_training.log_pyspark_test_set(
                features=features,
                predictions=predictions,
                labels=labels,
                raw_inputs=raw_inputs,
            )



# Note: Don't import anything from aporia.training here, it won't work without extra dependencies
import aporia.experimental  # noqa: F401
from .core.core_api import create_model, create_model_version, delete_model, init, shutdown
from .core.model_tags import add_model_tags, delete_model_tag, get_model_tags
from .core.types.model import ModelColor, ModelIcon
from .core.utils import generate_features_schema_from_shape
from .inference.api.dataset import SparkSQLDataSource, BigQueryDataSource
from .model import Model

__all__ = [
    # Core
    "create_model",
    "create_model_version",
    "delete_model",
    "init",
    "shutdown",
    "add_model_tags",
    "delete_model_tag",
    "get_model_tags",
    "ModelColor",
    "ModelIcon",
    "generate_features_schema_from_shape",
    # Inference
    "Model",
    "SparkSQLDataSource",
    "BigQueryDataSource"
]



import atexit
from contextlib import suppress
from functools import wraps
from http import HTTPStatus
import logging
from typing import Any, Callable, Dict, List, Optional, Union
import warnings

from aporia.model import Model
from .api.create_model_version import run_create_model_version_query
from .config import Config
from .context import get_context, init_context, reset_context
from .errors import AporiaError, AporiaHTTPError, handle_error
from .event_loop import EventLoop
from .http_client import HttpClient
from .logging_utils import init_logger, LOGGER_NAME, set_log_level
from .types.model import ModelColor, ModelIcon

logger = logging.getLogger(LOGGER_NAME)


def safe_api_function(error_message_format: str) -> Callable:
    """Decorator that wraps a function with error handling based on global configuration.

    Args:
        error_message_format: Error message format string - will be passed to handle_error()

    Returns:
        Wrapped function
    """

    def decorator_safe_api_function(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            context = get_context()
            try:
                return func(*args, **kwargs)
            except Exception as err:
                handle_error(
                    message_format=error_message_format,
                    verbose=context.config.verbose,
                    throw_errors=context.config.throw_errors,
                    debug=context.config.debug,
                    original_exception=err,
                )

            return None

        return wrapper

    return decorator_safe_api_function


def init(
    token: Optional[str] = None,
    host: Optional[str] = None,
    environment: Optional[str] = None,
    port: Optional[int] = None,
    verbose: Optional[bool] = None,
    throw_errors: Optional[bool] = None,
    debug: Optional[bool] = None,
    http_timeout_seconds: Optional[int] = None,
    verify_ssl: bool = True,
):
    """Initialize the Aporia SDK.

    Args:
        token: Authentication token.
        host: Controller host.
        environment: Environment in which aporia is initialized (e.g production, staging).
        port: Controller port. Defaults to 443.
        verbose: True to enable verbose error messages. Defaults to False.
        throw_errors: True to cause errors to be raised as exceptions. Defaults to False.
        debug: True to enable debug logs and stack traces in log messages. Defaults to False.
        http_timeout_seconds: HTTP timeout in seconds. Defaults to 30.

    Notes:
        * The token, host and environment parameters are required.
        * All of the parameters here can also be defined as environment variables:
            * token -> APORIA_TOKEN
            * host -> APORIA_HOST
            * environment -> APORIA_ENVIRONMENT
            * port -> APORIA_PORT
            * verbose -> APORIA_VERBOSE
            * throw_errors -> APORIA_THROW_ERRORS
            * debug -> APORIA_DEBUG
            * http_timeout_seconds -> APORIA_HTTP_TIMEOUT_SECONDS
        * Values passed as parameters to aporia.init() override the values from
          the corresponding environment variables.
    """
    init_logger()

    logger.debug("Initializing Aporia SDK.")

    try:
        config = Config(
            token=token,
            host=host,
            environment=environment,
            port=port,
            verbose=verbose,
            throw_errors=throw_errors,
            debug=debug,
            http_timeout_seconds=http_timeout_seconds,
            verify_ssl=verify_ssl,
        )

        set_log_level(config.debug)

        event_loop = EventLoop()
        http_client = HttpClient(
            token=config.token,
            host=config.host,
            port=config.port,
            default_timeout=config.http_timeout_seconds,
            verify_ssl=verify_ssl,
        )

        init_context(http_client=http_client, event_loop=event_loop, config=config)

        atexit.register(shutdown)
        logger.debug("Aporia SDK initialized.")
    except Exception as err:
        handle_error(
            message_format="Initializing Aporia SDK failed, error: {}",
            verbose=False if verbose is None else verbose,
            throw_errors=False if throw_errors is None else throw_errors,
            debug=False if debug is None else debug,
            original_exception=err,
        )


def shutdown():
    """Shuts down the Aporia SDK.

    Notes:
        * It is advised to call flush() before calling shutdown(), to ensure that
          all of the data that was sent reaches the controller.
    """
    logger.debug("Shutting down Aporia SDK.")
    with suppress(Exception):
        reset_context()


@safe_api_function("Creating model failed, error: {}")
def create_model(
    model_id: str,
    name: str,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    color: Optional[Union[ModelColor, str]] = None,
    icon: Optional[Union[ModelIcon, str]] = None,
    tags: Dict[str, str] = None,
) -> str:
    """Creates a new model.

    Args:
        model_id: A unique identifier for the new model, which will be used in all future operations
        name: A name for the new model, which will be displayed in Aporia's dashboard
        description: A description of the model
        owner: The email of the model owner
        color: A color to distinguish the model in Aporia's dashboard. Defaults to blue
        icon: An icon that indicates the model's designation. Defaults to general
        tags: A mapping of tag keys to tag values

    Returns:
        Model ID.

    Notes:
        * If this model_id already exists, NO EXCEPTION WILL BE RAISED!
          Instead, the same model ID will be returned.
    """
    context = get_context()

    try:
        response = context.event_loop.run_coroutine(
            context.http_client.post(
                url="/models",
                data={
                    "id": model_id,
                    "name": name,
                    "description": description,
                    "owner": owner,
                    "color": ModelColor(color).value if color is not None else ModelColor.BLUE,
                    "icon": ModelIcon(icon).value if icon is not None else ModelIcon.GENERAL,
                    "tags": tags,
                },
            )
        )

        return response["id"]
    except AporiaHTTPError as err:
        if err.http_status == HTTPStatus.CONFLICT:
            # This model ID already exists. We don't raise an exeption in this case.
            return model_id

        raise


@safe_api_function("Deleting model failed, error: {}")
def delete_model(model_id: str):
    """Deletes a model.

    Args:
        model_id: ID of the model to delete
    """
    context = get_context()
    context.event_loop.run_coroutine(context.http_client.delete(url=f"/models/{model_id}"))


@safe_api_function("Creating model version failed, error: {}")
def create_model_version(
    model_id: str,
    model_version: str,
    model_type: str,
    features: Dict[str, str],
    predictions: Dict[str, str],
    raw_inputs: Optional[Dict[str, str]] = None,
    metrics: Optional[Dict[str, str]] = None,
    model_data_type: Optional[str] = None,
    labels: Optional[List[str]] = None,
    multiclass_labels: Optional[List[str]] = None,
    feature_importance: Optional[Dict[str, float]] = None,
    mapping: Optional[Dict[str, str]] = None,
) -> Optional[Model]:
    """Creates a new model version, and defines a schema for it.

    Args:
        model_id: Model identifier, as received from the Aporia dashboard.
        model_version: Model version - this can be any string that represents the model
            version, such as "v1" or a git commit hash.
        model_type: Model type (also known as objective - see notes).
        features: Schema for model features (See notes).
        predictions: Schema for prediction results (See notes).
        raw_inputs: Schema for raw inputs (See notes).
        metrics: Schema for prediction metrics (See notes).
        model_data_type: Model data type.
        labels: Labels of multi-label, multiclass or binary model. Deprecated.
        multiclass_labels: Labels of multi-label, multiclass or binary model. Same as "labels", Deprecated.
        feature_importance: Features' importance.
        mapping: General mapping (See notes).

    Notes:
        * A schema is a dict, in which the keys are the fields you wish to report, and the
          values are the types of those fields. For example:
          ```python
            {
                "feature1": "numeric",
                "feature2": "datetime"
            }
          ```
        * The supported model types are:
            * "regression" - for regression models
            * "binary" - for binary classification models
            * "multiclass" - for multiclass classification models
            * "multi-label" - for multi-label classification models
            * "ranking" - for ranking models
        * The valid field types (and corresponding python types) are:

            | Field Type    | Python Types
            | ------------- | ------------
            | "numeric"     | float, int
            | "categorical" | int
            | "boolean"     | bool
            | "string"      | str
            | "datetime"    | datetime.datetime, or str representing a datetime in ISO-8601 format
            | "vector"      | list of floats
            | "text"        | str (to be used as free text)
            | "dict"        | dict[str, int]
        * The supported data types are:
            * "tabular"
            * "nlp"
        * The feature_importance is a mapping from feature name to it's importance (float).
          For example:
          ```python
             {
                "feature1": 1,
                "feature2": 2
             }
          ```
        * The mapping allowed fields are:
            * batch_id_column_name: The name of the key in the `raw_inputs`
            dict that holds the value of the `batch_id`.
            * relevance_column_name: The name of the key in the `predictions`
            dict that holds the value of the `relevance` score of models of type `ranking`.
            * actual_relevance_column_name: The name of the key in the `actuals`
            dict that holds the value of the `relevance` score of models of type `ranking`.

    Returns:
        Model object for the new version.
    """
    context = get_context()

    if len(model_id) == 0 or len(model_version) == 0:
        raise AporiaError("model_id and model_version must be non-empty strings.")

    if multiclass_labels is not None:
        warnings.warn(
            "The `multiclass_labels` parameter is deprecated, please use the `dict` field type.",
            DeprecationWarning,
            stacklevel=3,
        )

    if labels is not None:
        warnings.warn(
            "The `labels` parameter is deprecated, please use the `dict` field type.",
            DeprecationWarning,
            stacklevel=3,
        )

    context.event_loop.run_coroutine(
        run_create_model_version_query(
            http_client=context.http_client,
            model_id=model_id,
            model_version=model_version,
            model_type=model_type,
            model_data_type=model_data_type,
            features=features,
            predictions=predictions,
            raw_inputs=raw_inputs,
            metrics=metrics,
            labels=labels if labels is not None else multiclass_labels,
            feature_importance=feature_importance,
            mapping=mapping,
        )
    )

    return Model(model_id=model_id, model_version=model_version)



import asyncio
from http import HTTPStatus
import logging
import re
import ssl
from typing import Optional
from urllib.parse import urljoin

import aiohttp
import certifi
import orjson
from tenacity import AsyncRetrying, retry_if_exception, stop_after_attempt, wait_exponential

from .errors import AporiaError, AporiaHTTPError
from .logging_utils import LOGGER_NAME
from .utils import orjson_serialize_default_handler

logger = logging.getLogger(LOGGER_NAME)


REQUEST_MAX_ATTEMPTS = 4
REQUEST_TIMEOUT = 10
REQUEST_RETRY_INITIAL_SLEEP = 2
DEFAULT_TIMEOUT_SEC = 30
GRAPHQL_API_PATH = "/v1/controller/graphql"
REST_API_BASE_PATH = "/v1beta"
HTTP_ERROR_MESSAGES_BY_STATUS = {
    HTTPStatus.UNAUTHORIZED: "Authentication failed, please check your token.",
    HTTPStatus.NOT_FOUND: "Aporia server not found, please check the 'host' and 'token' parameters you passed to aporia.init()",  # noqa: B950
    HTTPStatus.METHOD_NOT_ALLOWED: "Communication with Aporia server failed, please check your token.",
}
HTTP_502_VERBOSE_ERROR = """The server couldn't handle the rate at which you are sending requests.
Please try one of the following solutions:
    * Reduce the amount of log_* calls per second
    * Upgrade your cluster so it can support a higher load.
"""


class HttpClient:
    """Asynchronous http client."""

    def __init__(self, token: str, host: str, port: int, default_timeout: int, verify_ssl: bool):
        """Initialize a HttpClient instance.

        Args:
            token: Authorization token
            host: Aporia server address
            port: Aporia server port
            default_timeout: Default timeout
            verify_ssl: Whether or not we should verify SSL certificates.
        """
        logger.debug("Initializing http client.")
        if not host.startswith("http"):
            host = f"https://{host}"

        self.base_url = f"{host}:{port}"
        if verify_ssl:
            self.ssl_context = ssl.create_default_context(cafile=certifi.where())
        else:
            self.ssl_context = False
            
        self.headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
        self.session = None
        self.default_timeout = default_timeout
        self.verify_ssl = verify_ssl

    async def open(self):
        """Opens the HTTP session."""
        logger.debug("Creating HTTP session with Aporia server.")
        self.session = aiohttp.ClientSession(
            headers=self.headers,
            connector=aiohttp.TCPConnector(ssl=False) if not self.verify_ssl else None,
            trust_env=True
        )

    async def graphql(
        self,
        query: str,
        variables: dict,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
        graphql_url: Optional[str] = None,
    ) -> dict:
        """Executes a GraphQL query and returns the result.

        Args:
            query: GraphQL query string
            variables: Variables for the query
            timeout: Timeout for the entire request, in seconds.
            retry_on_failure: True if the request should be retried on failure
                due to timeout or connection issues.
            graphql_url: GraphQL API URL.

        Returns:
            GraphQL query result
        """
        logger.debug("Sending GraphQL query", extra={"query": query, "variables": variables})

        data = {"query": query, "variables": variables}
        result = await self.request(
            method="POST",
            url=GRAPHQL_API_PATH if graphql_url is None else graphql_url,
            data=data,
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

        # This can only happen if we send a graphql request to something that doesn't implement
        # a GraphQL API - this check is mostly here to make the linter happy
        if result is None:
            raise AporiaError("Unexpected error, GraphQL API response was empty.")

        # GraphQL always returns 200, making difficult to catch GraphQL errors in the
        # generic error handling logic - we handle it here explicitly instead
        if "errors" in result:
            self._handle_http_error(status=HTTPStatus.BAD_REQUEST, content=result)

        return result["data"]

    async def delete(
        self,
        url: str,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs a DELETE HTTP request (see HttpClient.request())."""
        return await self.request(
            method="DELETE",
            url=f"{REST_API_BASE_PATH}/{url}",
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

    async def get(
        self,
        url: str,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs a GET HTTP request (see HttpClient.request())."""
        return await self.request(
            method="GET",
            url=f"{REST_API_BASE_PATH}/{url}",
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

    async def post(
        self,
        url: str,
        data: dict,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs a POST HTTP request (see HttpClient.request())."""
        return await self.request(
            method="POST",
            url=f"{REST_API_BASE_PATH}/{url}",
            data=data,
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

    async def request(  # type: ignore
        self,
        method: str,
        url: str,
        data: Optional[dict] = None,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs an HTTP request.

        Args:
            method: Request method string ("GET", "POST", etc.).
            url: Request URL.
            data: Optional request data, must be JSON serializable.
            timeout: Request timeout.
            retry_on_failure: True if the request should be retried on failure
                due to timeout or connection issues.

        Returns:
            The content of the HTTP response.
        """
        num_attempts = REQUEST_MAX_ATTEMPTS if retry_on_failure else 1
        timeout = timeout if timeout is not None else self.default_timeout
        try:
            async for attempt in AsyncRetrying(
                stop=stop_after_attempt(num_attempts),
                wait=wait_exponential(min=REQUEST_RETRY_INITIAL_SLEEP),
                retry=retry_if_exception(lambda err: not isinstance(err, AporiaError)),
                reraise=True,
            ):
                with attempt:
                    return await self._request(method=method, url=url, data=data, timeout=timeout)
        except asyncio.TimeoutError:
            raise AporiaError("Request timed out")

    async def _request(
        self, method: str, url: str, data: Optional[dict] = None, timeout: Optional[int] = None
    ) -> Optional[dict]:
        logger.debug(f"Sending {method} request to {url}.")

        # Strip redundant '/' from the URL
        normalized_url = re.sub("//+", "/", url)

        # Note: By default, aiohttp uses SSL and verifies the certificate in HTTPS requests
        async with self.session.request(  # type: ignore
            method=method,
            url=urljoin(self.base_url, normalized_url),
            data=None if data is None else self._serialize(data),
            timeout=timeout if timeout is not None else self.default_timeout,
            ssl=self.ssl_context,
        ) as response:
            response_content = await response.json()

            # All 2XX status codes (OK, CREATED, etc.) are considered success responses.
            # 3XX status codes are considered failures because we don't really handle redirects,
            # proxies, etc. All other statuses are obviously errors.
            if response.status < 200 or response.status >= 300:
                # This always raises an exception
                self._handle_http_error(
                    status=HTTPStatus(response.status), content=response_content
                )

            return response_content

    @staticmethod
    def _handle_http_error(status: HTTPStatus, content: Optional[dict] = None):
        short_message = None
        verbose_message = None

        if content is not None:
            # GraphQL error format
            if "errors" in content:
                short_message = content["errors"][0]["message"]
                verbose_message = content["errors"][0].get("extensions", {}).get("verbose_message")
            # REST API error format
            elif "error" in content:
                short_message = content["error"]["message"]
                verbose_message = content["error"].get("detail")

        # If error details were provided, raise an exception using those details
        # This should cover all errors that originate from our server.
        if short_message is not None:
            raise AporiaHTTPError(
                short_message=short_message,
                verbose_message=verbose_message,
                http_status=status,
            )

        # Otherwise, use a generic message for each exception type
        # This should cover authentication and routing errors.
        elif status in HTTP_ERROR_MESSAGES_BY_STATUS:
            raise AporiaHTTPError(
                short_message=HTTP_ERROR_MESSAGES_BY_STATUS[status],
                http_status=status,
            )

        # Bad gateway requires a more verbose error message, so we handle it separately
        elif status == HTTPStatus.BAD_GATEWAY:
            raise AporiaHTTPError(
                short_message="HTTP Gateway received invalid response from server",
                verbose_message=HTTP_502_VERBOSE_ERROR,
                http_status=status,
            )
        else:
            raise AporiaHTTPError(
                short_message=f"Unexpected HTTP error {status}",
                http_status=status,
            )

    @staticmethod
    def _serialize(data: dict) -> bytes:
        try:
            return orjson.dumps(
                data,
                option=orjson.OPT_SERIALIZE_NUMPY,
                default=orjson_serialize_default_handler,
            )
        except TypeError as err:
            raise AporiaError(str(err))

    async def close(self):
        """Closes the http session."""
        await self.session.close()



from collections import OrderedDict
from typing import Any, Tuple

import numpy as np

from .types.field import FieldType


def orjson_serialize_default_handler(obj: Any) -> str:
    """Default callback for orjson dumps.

    Supports pandas Timestamp type.

    Args:
        obj: Object to serialize

    Returns:
        Serialized object
    """
    # This supports serializing pandas the Timestamp type
    if hasattr(obj, "isoformat"):
        return obj.isoformat()

    # As recommended by orjson, we raise TypeError to indicate this
    # function can't serialize the object
    raise TypeError()


def generate_features_schema_from_shape(features_shape: Tuple) -> OrderedDict:
    """Return features schema to use in version creation from ndarray shape.

    Args:
        features_shape (Tuple): the shape of the features array.

    Returns:
        OrderedDict: The object to pass to features in the version schema.
    """
    features_schema = OrderedDict()
    for i in range(int(np.prod(features_shape))):
        features_schema[f"X{i}"] = FieldType.NUMERIC.value

    return features_schema



import atexit
from concurrent.futures import ALL_COMPLETED, wait
from datetime import datetime, timezone
from functools import partial
import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from numpy import ndarray

from aporia.core.base_model import BaseModel, validate_model_ready
from aporia.core.errors import handle_error
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldValue
from aporia.inference.api.dataset import connect_dataset, DataSource
from aporia.inference.api.log_model_properties import (
    log_index_to_word_mapping,
    set_feature_importance,
    upload_model_artifact,
)
from .api.inference_fragment import ActualsFragment, Fragment, PredictionFragment, RawInputsFragment
from .api.log_inference import log_inference_fragments
from .api.log_json import log_json
from .inference_queue import InferenceQueue
from .types.model_artifact import ModelArtifactType

logger = logging.getLogger(LOGGER_NAME)


class InferenceModel(BaseModel):
    """Model object for logging inference events."""

    def __init__(self, model_id: str, model_version: str):
        """Initializes an inference model object.

        Args:
            model_id: Model identifier, as received from the Aporia dashboard.
            model_version: Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
        """
        try:
            super().__init__(
                model_id=model_id, model_version=model_version, set_ready_when_done=False
            )

            # We keep a list of all tasks that were not awaited, to allow flushing
            # We have to do this manually to support python versions below
            # 3.7 (otherwise we could use asyncio.all_tasks())
            self._futures = []  # type: ignore
            self._queue = InferenceQueue(
                event_loop=self._event_loop,
                batch_size=self._config.queue_batch_size,
                max_size=self._config.queue_max_size,
                flush_interval=self._config.queue_flush_interval,
                flush_callback=self._flush_inference_callback,
            )

            self._model_ready = True

            if self._config.verbose:
                atexit.register(self._warn_about_unfinished_tasks)

        except Exception as err:
            config = getattr(self, "_config", None)
            handle_error(
                message_format="Model initialization failed, error: {}",
                verbose=False if config is None else config.verbose,
                throw_errors=False if config is None else config.throw_errors,
                debug=False if config is None else config.debug,
                log_level=logging.ERROR,
                original_exception=err,
            )

    @validate_model_ready
    def log_prediction(
        self,
        id: str,
        features: Union[Dict[str, FieldValue], ndarray],
        predictions: Dict[str, FieldValue],
        metrics: Optional[Dict[str, FieldValue]] = None,
        occurred_at: Optional[datetime] = None,
        confidence: Optional[Union[float, List[float]]] = None,
        raw_inputs: Optional[Dict[str, FieldValue]] = None,
        actuals: Optional[Dict[str, FieldValue]] = None,
    ):
        """Logs a single prediction.

        Args:
            id: Prediction identifier.
            features: Values for all the features in the prediction
            predictions: Prediction result
            metrics: Prediction metrics.
            occurred_at: Prediction timestamp.
            confidence: Prediction confidence.
            raw_inputs: Raw inputs of the prediction.
            actuals: Actual prediction results.

        Note:
            * If occurred_at is None, it will be reported as datetime.now()
        """
        self.log_batch_prediction(
            batch_predictions=[
                dict(
                    id=id,
                    features=features,
                    predictions=predictions,
                    metrics=metrics,
                    occurred_at=occurred_at,
                    confidence=confidence,
                    raw_inputs=raw_inputs,
                    actuals=actuals,
                )
            ]
        )

    @validate_model_ready
    def log_batch_prediction(self, batch_predictions: Iterable[dict]):
        """Logs multiple predictions.

        Args:
            batch_predictions: An iterable that produces prediction dicts.

                * Each prediction dict MUST contain the following keys:
                    * id (str): Prediction identifier.
                    * features (Dict[str, FieldValue]): Values for all the features in the prediction
                    * predictions (Dict[str, FieldValue]): Prediction result

                * Each prediction dict MAY also contain the following keys:
                    * occurred_at (datetime): Prediction timestamp.
                    * metrics (Dict[str, FieldValue]): Prediction metrics
                    * confidence (Union[float, List[float]]): Prediction confidence.
                    * raw_inputs (Dict[str, FieldValue]): Raw inputs of the prediction.
                    * actuals (Dict[str, FieldValue]) Actual prediction results.

        Notes:
            * If occurred_at is None in any of the predictions, it will be reported as datetime.now()
        """
        self._log_batch_inference(
            batch=batch_predictions,
            fragment_class=partial(PredictionFragment, timestamp=datetime.now(tz=timezone.utc)),
            error_message="Logging prediction batch failed, error: {}",
        )

    @validate_model_ready
    def log_raw_inputs(self, id: str, raw_inputs: Dict[str, FieldValue]):
        """Logs raw inputs of a single prediction.

        Args:
            id: Prediction identifier.
            raw_inputs: Raw inputs of the prediction.
        """
        self.log_batch_raw_inputs(batch_raw_inputs=[dict(id=id, raw_inputs=raw_inputs)])

    @validate_model_ready
    def log_batch_raw_inputs(self, batch_raw_inputs: Iterable[dict]):
        """Logs raw inputs of multiple predictions.

        Args:
            batch_raw_inputs: An iterable that produces raw_inputs dicts.

                * Each dict MUST contain the following keys:
                    * id (str): Prediction identifier.
                    * raw_inputs (Dict[str, FieldValue]): Raw inputs of the prediction.
        """
        self._log_batch_inference(
            batch=batch_raw_inputs,
            fragment_class=RawInputsFragment,
            error_message="Logging raw inputs batch failed, error: {}",
        )

    @validate_model_ready
    def log_actuals(self, id: str, actuals: Dict[str, FieldValue]):
        """Logs actual values of a single prediction.

        Args:
            id: Prediction identifier.
            actuals: Actual prediction results.

        Note:
            * The fields reported in actuals must be a subset of the fields reported
              in predictions.
        """
        self.log_batch_actuals(batch_actuals=[dict(id=id, actuals=actuals)])

    @validate_model_ready
    def log_batch_actuals(self, batch_actuals: Iterable[dict]):
        """Logs actual values of multiple predictions.

        Args:
            batch_actuals: An iterable that produces actuals dicts.

                * Each dict MUST contain the following keys:
                    * id (str): Prediction identifier.
                    * actuals (Dict[str, FieldValue]): Actual prediction results.

        Note:
            * The fields reported in actuals must be a subset of the fields reported
              in predictions.
        """
        self._log_batch_inference(
            batch=batch_actuals,
            fragment_class=ActualsFragment,
            error_message="Logging actuals batch failed, error: {}",
        )

    def _log_batch_inference(
        self,
        batch: Iterable[dict],
        fragment_class: Callable,
        error_message: str,
    ):
        with self.handle_error(error_message):
            fragments = [fragment_class(data_point) for data_point in batch]

            if self._config.verbose:
                for i, fragment in enumerate(fragments):
                    if not fragment.is_valid():
                        logger.warning("{} {} is not valid".format(fragment.type.value, i))

            count = self._event_loop.run_coroutine(self._queue.put(fragments=fragments))

            dropped_fragments = len(fragments) - count
            if dropped_fragments > 0:
                logger.warning(
                    "Message queue reached size limit, dropped {} messages.".format(
                        dropped_fragments
                    )
                )

    async def _flush_inference_callback(self, fragments: List[Fragment]):
        with self.handle_error("Server error: {}", throw_errors=False):
            serialized_fragments = []
            for fragment in fragments:
                try:
                    serialized_fragments.append(fragment.serialize())
                except Exception as err:
                    logger.error("Serializing data failed, error: {}".format(err))

            if len(serialized_fragments) > 0:
                await log_inference_fragments(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    environment=self._config.environment,
                    serialized_fragments=serialized_fragments,
                    await_insert=self._config.verbose,
                )

    @validate_model_ready
    def log_json(self, data: Any):
        """Logs arbitrary data.

        Args:
            data: Data to log, must be JSON serializable
        """
        logger.debug("Logging arbitrary data.")
        with self.handle_error("Logging arbitrary data failed, error: {}"):
            future = self._event_loop.create_task(self._log_json(data=data))
            self._futures.append(future)

    async def _log_json(self, data: Any):
        with self.handle_error("Logging arbitrary data failed, error: {}", throw_errors=False):
            await log_json(
                http_client=self._http_client,
                model_id=self.model_id,
                model_version=self.model_version,
                environment=self._config.environment,
                data=data,
            )

    @validate_model_ready
    def upload_model_artifact(
        self, model_artifact: Any, artifact_type: Union[str, ModelArtifactType]
    ):
        """Uploads binary model artifact.

        Args:
            model_artifact: Binary model artifact.
            artifact_type: The type of model artifact (see below)

        Model Artifact Types:
            - onnx
            - h5
        """
        with self.handle_error("Uploading model artifact failed, error: {}"):
            artifact_type = ModelArtifactType(artifact_type)

            self._event_loop.run_coroutine(
                upload_model_artifact(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    model_artifact=model_artifact,
                    artifact_type=artifact_type,
                )
            )

    @validate_model_ready
    def set_feature_importance(self, feature_importance: Dict[str, float]):
        """Update the features' importance of the model.

        Args:
            feature_importance: feature name to importance mapping
        """
        with self.handle_error("Setting feature importance failed, error: {}", throw_errors=False):
            self._event_loop.run_coroutine(
                set_feature_importance(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    feature_importance=feature_importance,
                )
            )

    @validate_model_ready
    def log_index_to_word_mapping(self, index_to_word_mapping: Dict[Any, Any]):
        """Logs index to word mapping.

        Args:
            index_to_word_mapping: A mapping between a numeric index to a word.
        """
        with self.handle_error(
            "Logging index to word mapping failed, error: {}", throw_errors=False
        ):
            self._event_loop.run_coroutine(
                log_index_to_word_mapping(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    index_to_word_mapping=index_to_word_mapping,
                )
            )

    @validate_model_ready
    def connect_serving(
        self,
        data_source: DataSource,
        id_column: str,
        timestamp_column: str,
        predictions: Optional[Dict[str, str]] = None,
        features: Optional[Dict[str, str]] = None,
        labels: Optional[Dict[str, str]] = None,
        raw_inputs: Optional[Dict[str, str]] = None,
    ):
        """Connect to external serving data set.

        Args:
            data_source: The data source to fetch the data set from.
            id_column: The name of the id column.
            timestamp_column: The name of the timestamp column.
            features: Mapping from feature name to column name.
            predictions: Mapping from prediction name to column name.
            labels: Mapping from actual name to column name.
            raw_inputs: Mapping from raw input name to column name.
        """
        with self.handle_error("Connection to serving failed, error: {}"):
            self._event_loop.run_coroutine(
                connect_dataset(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    environment=self._config.environment,
                    dataset="serving",
                    data_source=data_source,
                    id_column=id_column,
                    timestamp_column=timestamp_column,
                    features=features,
                    predictions=predictions,
                    labels=labels,
                    raw_inputs=raw_inputs,
                )
            )

    @validate_model_ready
    def connect_training(
        self,
        data_source: DataSource,
        id_column: str,
        timestamp_column: str,
        predictions: Optional[Dict[str, str]] = None,
        features: Optional[Dict[str, str]] = None,
        labels: Optional[Dict[str, str]] = None,
        raw_inputs: Optional[Dict[str, str]] = None,
    ):
        """Connect to external training data set.

        Args:
            data_source: The data source to fetch the data set from.
            id_column: The name of the id column.
            timestamp_column: The name of the timestamp column.
            features: Mapping from feature name to column name.
            predictions: Mapping from prediction name to column name.
            labels: Mapping from prediction name to column name.
            raw_inputs: Mapping from raw input name to column name.
        """
        with self.handle_error("Connection to training set failed, error: {}"):
            self._event_loop.run_coroutine(
                connect_dataset(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    environment=self._config.environment,
                    dataset="training",
                    data_source=data_source,
                    id_column=id_column,
                    timestamp_column=timestamp_column,
                    features=features,
                    predictions=predictions,
                    labels=labels,
                    raw_inputs=raw_inputs,
                )
            )

    @validate_model_ready
    def connect_testing(
        self,
        data_source: DataSource,
        id_column: str,
        timestamp_column: str,
        predictions: Optional[Dict[str, str]] = None,
        features: Optional[Dict[str, str]] = None,
        labels: Optional[Dict[str, str]] = None,
        raw_inputs: Optional[Dict[str, str]] = None,
    ):
        """Connect to external test data set.

        Args:
            data_source: The data source to fetch the data set from.
            id_column: The name of the id column.
            timestamp_column: The name of the timestamp column.
            features: Mapping from feature name to column name.
            predictions: Mapping from prediction name to column name.
            labels: Mapping from prediction name to column name.
            raw_inputs: Mapping from raw input name to column name.
        """
        with self.handle_error("Connection to test set failed, error: {}"):
            self._event_loop.run_coroutine(
                connect_dataset(
                    http_client=self._http_client,
                    model_id=self.model_id,
                    model_version=self.model_version,
                    environment=self._config.environment,
                    dataset="testing",
                    data_source=data_source,
                    id_column=id_column,
                    timestamp_column=timestamp_column,
                    features=features,
                    predictions=predictions,
                    labels=labels,
                    raw_inputs=raw_inputs,
                )
            )

    @validate_model_ready
    def flush(self, timeout: Optional[int] = None) -> Optional[int]:
        """Waits for all currently scheduled tasks to finish.

        Args:
            timeout: Maximum number of seconds to wait for tasks to
                complete. Default to None (No timeout).

        Returns:
            Number of tasks that haven't finished running.
        """
        with self.handle_error("Flushing remaining data failed, error: {}"):
            futures = self._futures
            self._futures = []

            logger.debug("Flushing predictions.")
            # Add a task that flushes the queue, and another that waits for the flush to complete
            futures.append(self._event_loop.create_task(self._queue.flush()))
            futures.append(self._event_loop.create_task(self._queue.join()))

            logger.debug("Waiting for {} scheduled tasks to finish executing.".format(len(futures)))
            done, not_done = wait(futures, timeout=timeout, return_when=ALL_COMPLETED)

            logger.debug(
                "{} tasks finished, {} tasks not finished.".format(len(done), len(not_done))
            )
            self._futures.extend(not_done)

            return len(not_done)

        return None

    def _warn_about_unfinished_tasks(self):
        if len(self._queue) > 0:
            logger.warning(
                "The process was closed before the SDK finished flushing all of the logged "
                "predictions, please call apr_model.flush() before the end of your "
                "program to ensure that all of the predictions have reached the server."
            )
